"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.start = void 0;
const fs_1 = require("fs");
const util_1 = require("util");
const path_1 = __importDefault(require("path"));
const cli_meta_1 = require("@pnpm/cli-meta");
const error_1 = require("@pnpm/error");
const logger_1 = require("@pnpm/logger");
const server_1 = require("@pnpm/server");
const store_connection_manager_1 = require("@pnpm/store-connection-manager");
const store_path_1 = require("@pnpm/store-path");
const diable_1 = __importDefault(require("@zkochan/diable"));
const get_port_1 = __importDefault(require("get-port"));
const is_windows_1 = __importDefault(require("is-windows"));
const signal_exit_1 = __importDefault(require("signal-exit"));
const storeServerLogger = (0, logger_1.logger)('store-server');
const write = (0, util_1.promisify)(fs_1.write);
const close = (0, util_1.promisify)(fs_1.close);
const open = (0, util_1.promisify)(fs_1.open);
async function start(opts) {
    if (opts.protocol === 'ipc' && opts.port) {
        throw new Error('Port cannot be selected when server communicates via IPC');
    }
    if (opts.background && !diable_1.default.isDaemon()) {
        (0, diable_1.default)();
    }
    const storeDir = await (0, store_path_1.getStorePath)({
        pkgRoot: opts.dir,
        storePath: opts.storeDir,
        pnpmHomeDir: opts.pnpmHomeDir,
    });
    const connectionInfoDir = (0, store_connection_manager_1.serverConnectionInfoDir)(storeDir);
    const serverJsonPath = path_1.default.join(connectionInfoDir, 'server.json');
    await fs_1.promises.mkdir(connectionInfoDir, { recursive: true });
    // Open server.json with exclusive write access to ensure only one process can successfully
    // start the server. Note: NFS does not support exclusive writing, but do we really care?
    // Source: https://github.com/moxystudio/node-proper-lockfile#user-content-comparison
    let fd;
    try {
        fd = await open(serverJsonPath, 'wx');
    }
    catch (error) { // eslint-disable-line
        if (error.code !== 'EEXIST') {
            throw error;
        }
        throw new error_1.PnpmError('SERVER_MANIFEST_LOCKED', `Canceling startup of server (pid ${process.pid}) because another process got exclusive access to server.json`);
    }
    let server = null;
    (0, signal_exit_1.default)(() => {
        if (server !== null) {
            // Note that server.close returns a Promise, but we cannot wait for it because we may be
            // inside the 'exit' even of process.
            server.close(); // eslint-disable-line @typescript-eslint/no-floating-promises
        }
        if (fd !== null) {
            try {
                (0, fs_1.closeSync)(fd);
            }
            catch (error) { // eslint-disable-line
                storeServerLogger.error(error, 'Got error while closing file descriptor of server.json, but the process is already exiting');
            }
        }
        try {
            (0, fs_1.unlinkSync)(serverJsonPath);
        }
        catch (error) { // eslint-disable-line
            if (error.code !== 'ENOENT') {
                storeServerLogger.error(error, 'Got error unlinking server.json, but the process is already exiting');
            }
        }
    });
    const store = await (0, store_connection_manager_1.createNewStoreController)(Object.assign(opts, {
        storeDir,
    }));
    const protocol = opts.protocol ?? (opts.port ? 'tcp' : 'auto');
    const serverOptions = await getServerOptions(connectionInfoDir, { protocol, port: opts.port });
    const connectionOptions = {
        remotePrefix: serverOptions.path != null
            ? `http://unix:${serverOptions.path}:`
            : `http://${serverOptions.hostname}:${serverOptions.port}`,
    };
    server = (0, server_1.createServer)(store.ctrl, {
        ...serverOptions,
        ignoreStopRequests: opts.ignoreStopRequests,
        ignoreUploadRequests: opts.ignoreUploadRequests,
    });
    // Make sure to populate server.json after the server has started, so clients know that the server is
    // listening if a server.json with valid JSON content exists.
    const serverJson = {
        connectionOptions,
        pid: process.pid,
        pnpmVersion: cli_meta_1.packageManager.version,
    };
    const serverJsonStr = JSON.stringify(serverJson, undefined, 2); // undefined and 2 are for formatting.
    const serverJsonBuffer = Buffer.from(serverJsonStr, 'utf8');
    // fs.write on NodeJS 4 requires the parameters offset and length to be set:
    // https://nodejs.org/docs/latest-v4.x/api/fs.html#fs_fs_write_fd_buffer_offset_length_position_callback
    await write(fd, serverJsonBuffer, 0, serverJsonBuffer.byteLength);
    const fdForClose = fd;
    // Set fd to null so we only attempt to close it once.
    fd = null;
    await close(fdForClose);
    // Intentionally avoid returning control back to the caller until the server
    // exits. This defers cleanup operations that should not run before the server
    // finishes.
    await server.waitForClose;
}
exports.start = start;
async function getServerOptions(connectionInfoDir, opts) {
    switch (opts.protocol) {
        case 'tcp':
            return getTcpOptions();
        case 'ipc':
            if ((0, is_windows_1.default)()) {
                throw new Error('IPC protocol is not supported on Windows currently');
            }
            return getIpcOptions();
        case 'auto':
            if ((0, is_windows_1.default)()) {
                return getTcpOptions();
            }
            return getIpcOptions();
        default:
            throw new Error(`Protocol ${opts.protocol} is not supported`);
    }
    async function getTcpOptions() {
        return {
            hostname: 'localhost',
            port: opts.port || await (0, get_port_1.default)({ port: 5813 }), // eslint-disable-line
        };
    }
    function getIpcOptions() {
        return {
            path: path_1.default.join(connectionInfoDir, 'socket'),
        };
    }
}
//# sourceMappingURL=start.js.map